@@ -88,7 +88,8 @@ static const wcstring &C_(const wcstring &s) { return s; }
 /// If option is non-empty, it specifies a switch for the command. If \c comp is also not empty, it
 /// contains a list of non-switch arguments that may only follow directly after the specified
 /// switch.
-using complete_entry_opt_t = struct complete_entry_opt {
+namespace {
+struct complete_entry_opt_t {
     // Text of the option (like 'foo').
     wcstring option;
     // Type of the option: args-oly, short, single_long, or double_long.
@@ -139,16 +140,33 @@ class completion_entry_t {
     const unsigned int order;
 
     /// Getters for option list.
-    const option_list_t &get_options() const;
+    const option_list_t &get_options() const { return options; }
 
     /// Adds or removes an option.
-    void add_option(const complete_entry_opt_t &opt);
+    void add_option(const complete_entry_opt_t &opt) { options.push_front(opt); }
     bool remove_option(const wcstring &option, complete_option_type_t type);
 
     completion_entry_t(wcstring c, bool type)
         : cmd(std::move(c)), cmd_is_path(type), order(++k_complete_order) {}
 };
 
+/// Remove all completion options in the specified entry that match the specified short / long
+/// option strings. Returns true if it is now empty and should be deleted, false if it's not empty.
+/// Must be called while locked.
+bool completion_entry_t::remove_option(const wcstring &option, complete_option_type_t type) {
+    auto iter = this->options.begin();
+    while (iter != this->options.end()) {
+        if (iter->option == option && iter->type == type) {
+            iter = this->options.erase(iter);
+        } else {
+            // Just go to the next one.
+            ++iter;
+        }
+    }
+    return this->options.empty();
+}
+}  // namespace
+
 /// Set of all completion entries.
 namespace std {
 template <>
@@ -164,6 +182,7 @@ struct equal_to<completion_entry_t> {
         return c1.cmd == c2.cmd;
     }
 };
+
 }  // namespace std
 using completion_entry_set_t = std::unordered_set<completion_entry_t>;
 static owning_lock<completion_entry_set_t> s_completion_set;
@@ -178,10 +197,6 @@ static bool compare_completions_by_order(const completion_entry_t &p1,
     return p1.order < p2.order;
 }
 
-void completion_entry_t::add_option(const complete_entry_opt_t &opt) { options.push_front(opt); }
-
-const option_list_t &completion_entry_t::get_options() const { return options; }
-
 description_func_t const_desc(const wcstring &s) {
     return [=](const wcstring &ignored) {
         UNUSED(ignored);
@@ -343,6 +358,7 @@ void completions_sort_and_prioritize(completion_list_t *comps, completion_reques
     }
 }
 
+namespace {
 /// Class representing an attempt to compute completions.
 class completer_t {
     /// The operation context for this completion.
@@ -461,12 +477,6 @@ class completer_t {
 // Autoloader for completions.
 static owning_lock<autoload_t> completion_autoloader{autoload_t(L"fish_complete_path")};
 
-/// Create a new completion entry.
-void append_completion(completion_list_t *completions, wcstring comp, wcstring desc,
-                       complete_flags_t flags, string_fuzzy_match_t match) {
-    completions->emplace_back(std::move(comp), std::move(desc), match, flags);
-}
-
 /// Test if the specified script returns zero. The result is cached, so that if multiple completions
 /// use the same condition, it needs only be evaluated once. condition_cache_clear must be called
 /// after a completion run to make sure that there are no stale completions.
@@ -479,7 +489,6 @@ bool completer_t::condition_test(const wcstring &condition) {
         return false;
     }
 
-    ASSERT_IS_MAIN_THREAD();
     bool test_res;
     auto cached_entry = condition_cache.find(condition);
     if (cached_entry == condition_cache.end()) {
@@ -505,77 +514,16 @@ static completion_entry_t &complete_get_exact_entry(completion_entry_set_t &comp
     return const_cast<completion_entry_t &>(*ins.first);
 }
 
-void complete_add(const wchar_t *cmd, bool cmd_is_path, const wcstring &option,
-                  complete_option_type_t option_type, completion_mode_t result_mode,
-                  const wchar_t *condition, const wchar_t *comp, const wchar_t *desc,
-                  complete_flags_t flags) {
-    assert(cmd && "Null command");
-    // option should be empty iff the option type is arguments only.
-    assert(option.empty() == (option_type == option_type_args_only));
-
-    // Lock the lock that allows us to edit the completion entry list.
-    auto completion_set = s_completion_set.acquire();
-    completion_entry_t &c = complete_get_exact_entry(*completion_set, cmd, cmd_is_path);
-
-    // Create our new option.
-    complete_entry_opt_t opt;
-    opt.option = option;
-    opt.type = option_type;
-    opt.result_mode = result_mode;
-
-    if (comp) opt.comp = comp;
-    if (condition) opt.condition = condition;
-    if (desc) opt.desc = desc;
-    opt.flags = flags;
-
-    c.add_option(opt);
-}
-
-/// Remove all completion options in the specified entry that match the specified short / long
-/// option strings. Returns true if it is now empty and should be deleted, false if it's not empty.
-/// Must be called while locked.
-bool completion_entry_t::remove_option(const wcstring &option, complete_option_type_t type) {
-    auto iter = this->options.begin();
-    while (iter != this->options.end()) {
-        if (iter->option == option && iter->type == type) {
-            iter = this->options.erase(iter);
-        } else {
-            // Just go to the next one.
-            ++iter;
-        }
-    }
-    return this->options.empty();
-}
-
-void complete_remove(const wcstring &cmd, bool cmd_is_path, const wcstring &option,
-                     complete_option_type_t type) {
-    auto completion_set = s_completion_set.acquire();
-
-    completion_entry_t tmp_entry(cmd, cmd_is_path);
-    auto iter = completion_set->find(tmp_entry);
-    if (iter != completion_set->end()) {
-        // const_cast: See SET_ELEMENTS_ARE_IMMUTABLE.
-        auto &entry = const_cast<completion_entry_t &>(*iter);
-
-        bool delete_it = entry.remove_option(option, type);
-        if (delete_it) {
-            completion_set->erase(iter);
-        }
-    }
-}
-
-void complete_remove_all(const wcstring &cmd, bool cmd_is_path) {
-    auto completion_set = s_completion_set.acquire();
-    completion_entry_t tmp_entry(cmd, cmd_is_path);
-    completion_set->erase(tmp_entry);
-}
-
 /// Find the full path and commandname from a command string 'str'.
 static void parse_cmd_string(const wcstring &str, wcstring *path, wcstring *cmd,
                              const environment_t &vars) {
-    if (!path_get_path(str, path, vars)) {
-        /// Use the empty string as the 'path' for commands that can not be found.
-        path->clear();
+    bool found = path_get_path(str, path, vars);
+    // If the command was not found, 'path' is the empty string.
+    // Resolve commands that use relative paths because we compare full paths with "complete -p".
+    if (found && !str.empty() && str.at(0) != L'/') {
+        if (auto full_path = wrealpath(*path)) {
+            path->assign(full_path.acquire());
+        }
     }
 
     // Make sure the path is not included in the command.
@@ -646,7 +594,7 @@ void completer_t::complete_cmd_desc(const wcstring &str) {
     // least two characters if we don't know the location of the whatis-database.
     if (cmd.length() < 2) return;
 
-    if (wildcard_has(cmd, false)) {
+    if (wildcard_has(cmd)) {
         return;
     }
 
@@ -698,7 +646,7 @@ void completer_t::complete_cmd_desc(const wcstring &str) {
         // And once again I make sure the first character is uppercased because I like it that
         // way, and I get to decide these things.
         val.at(0) = towupper(val.at(0));
-        lookup.insert(std::make_pair(std::move(key), std::move(val)));
+        lookup.emplace(std::move(key), std::move(val));
     }
 
     // Then do a lookup on every completion and if a match is found, change to the new
@@ -712,16 +660,10 @@ void completer_t::complete_cmd_desc(const wcstring &str) {
 
 /// Returns a description for the specified function, or an empty string if none.
 static wcstring complete_function_desc(const wcstring &fn) {
-    wcstring result;
-    bool has_description = function_get_desc(fn, result);
-    if (!has_description) {
-        function_get_definition(fn, result);
-        // A completion description is a single line.
-        for (wchar_t &c : result) {
-            if (c == L'\n') c = L' ';
-        }
+    if (auto props = function_get_props(fn)) {
+        return props->description;
     }
-    return result;
+    return wcstring{};
 }
 
 /// Complete the specified command name. Search for executables in the path, executables defined
@@ -913,7 +855,7 @@ static void complete_load(const wcstring &name) {
     // We have to load this as a function, since it may define a --wraps or signature.
     // See issue #2466.
     auto &parser = parser_t::principal_parser();
-    function_load(name, parser);
+    function_get_props_autoload(name, parser);
 
     // It's important to NOT hold the lock around completion loading.
     // We need to take the lock to decide what to load, drop it to perform the load, then reacquire
@@ -1067,6 +1009,7 @@ bool completer_t::complete_param_for_command(const wcstring &cmd_orig, const wcs
             if (!this->condition_test(o.condition)) continue;
             if (o.option.empty()) {
                 use_files = use_files && (!(o.result_mode.no_files));
+                has_force = has_force || o.result_mode.force_files;
                 complete_from_args(str, o.comp, o.localized_desc(), o.flags);
             }
 
@@ -1151,7 +1094,9 @@ bool completer_t::complete_param_for_command(const wcstring &cmd_orig, const wcs
         }
     }
 
-    if (!(has_force || use_files)) {
+    if (has_force) {
+        *out_do_file = true;
+    } else if (!use_files) {
         *out_do_file = false;
     }
     return true;
@@ -1445,7 +1390,7 @@ void completer_t::complete_custom(const wcstring &cmd, const wcstring &cmdline,
     // buitin_commandline will refer to the wrapped command. But not if
     // we're doing autosuggestions.
     maybe_t<cleanup_t> remove_transient{};
-    bool wants_transient = ad->wrap_depth > 0 && !is_autosuggest;
+    bool wants_transient = (ad->wrap_depth > 0 || ad->var_assignments) && !is_autosuggest;
     if (wants_transient) {
         ctx.parser->libdata().transient_commandlines.push_back(cmdline);
         remove_transient.emplace([=] { ctx.parser->libdata().transient_commandlines.pop_back(); });
@@ -1461,6 +1406,14 @@ void completer_t::complete_custom(const wcstring &cmd, const wcstring &cmdline,
     }
 }
 
+static bool expand_command_token(const operation_context_t &ctx, wcstring &cmd_tok) {
+    // TODO: we give up if the first token expands to more than one argument. We could handle
+    // that case by propagating arguments.
+    // Also we could expand wildcards.
+    return expand_one(cmd_tok, {expand_flag::skip_cmdsubst, expand_flag::skip_wildcards}, ctx,
+                      nullptr);
+}
+
 // Invoke command-specific completions given by \p arg_data.
 // Then, for each target wrapped by the given command, update the command
 // line with that target and invoke this recursively.
@@ -1500,6 +1453,7 @@ void completer_t::walk_wrap_chain(const wcstring &cmd, const wcstring &cmdline,
             } else {
                 wrapped_command_offset_in_wt = tok->offset;
                 wrapped_command = std::move(tok_src);
+                expand_command_token(ctx, wrapped_command);
                 break;
             }
         }
@@ -1621,6 +1575,7 @@ void completer_t::perform_for_commandline(wcstring cmdline) {
     // Get all the arguments.
     std::vector<tok_t> tokens;
     parse_util_process_extent(cmdline.c_str(), position_in_statement, nullptr, nullptr, &tokens);
+    size_t actual_token_count = tokens.size();
 
     // Hack: fix autosuggestion by removing prefixing "and"s #6249.
     if (is_autosuggest) {
@@ -1649,6 +1604,14 @@ void completer_t::perform_for_commandline(wcstring cmdline) {
         return;
     }
 
+    wcstring *effective_cmdline, effective_cmdline_buf;
+    if (tokens.size() == actual_token_count) {
+        effective_cmdline = &cmdline;
+    } else {
+        effective_cmdline_buf.assign(cmdline, tokens.front().offset, wcstring::npos);
+        effective_cmdline = &effective_cmdline_buf;
+    }
+
     const tok_t &cmd_tok = tokens.front();
     const tok_t &cur_tok = tokens.back();
 
@@ -1720,20 +1683,18 @@ void completer_t::perform_for_commandline(wcstring cmdline) {
         custom_arg_data_t arg_data{&var_assignments};
         arg_data.had_ddash = had_ddash;
 
-        assert(cmd_tok.offset < std::numeric_limits<uint32_t>::max());
-        assert(cmd_tok.length < std::numeric_limits<uint32_t>::max());
-        source_range_t command_range = {static_cast<uint32_t>(cmd_tok.offset),
-                                        static_cast<uint32_t>(cmd_tok.length)};
+        source_offset_t bias = cmdline.size() - effective_cmdline->size();
+        source_range_t command_range = {cmd_tok.offset - bias, cmd_tok.length};
 
-        wcstring unesc_command;
+        wcstring exp_command = cmd_tok.get_source(cmdline);
         bool unescaped =
-            unescape_string(cmd_tok.get_source(cmdline), &unesc_command, UNESCAPE_DEFAULT) &&
+            expand_command_token(ctx, exp_command) &&
             unescape_string(previous_argument, &arg_data.previous_argument, UNESCAPE_DEFAULT) &&
             unescape_string(current_argument, &arg_data.current_argument, UNESCAPE_INCOMPLETE);
         if (unescaped) {
             // Have to walk over the command and its entire wrap chain. If any command
             // disables do_file, then they all do.
-            walk_wrap_chain(unesc_command, cmdline, command_range, &arg_data);
+            walk_wrap_chain(exp_command, *effective_cmdline, command_range, &arg_data);
             do_file = arg_data.do_file;
 
             // If we're autosuggesting, and the token is empty, don't do file suggestions.
@@ -1743,7 +1704,8 @@ void completer_t::perform_for_commandline(wcstring cmdline) {
         }
 
         // Hack. If we're cd, handle it specially (issue #1059, others).
-        handle_as_special_cd = (unesc_command == L"cd");
+        handle_as_special_cd =
+            (exp_command == L"cd") || arg_data.visited_wrapped_commands.count(L"cd");
     }
 
     // Maybe apply variable assignments.
@@ -1760,6 +1722,63 @@ void completer_t::perform_for_commandline(wcstring cmdline) {
     mark_completions_duplicating_arguments(cmdline, current_token, tokens);
 }
 
+}  // namespace
+
+/// Create a new completion entry.
+void append_completion(completion_list_t *completions, wcstring comp, wcstring desc,
+                       complete_flags_t flags, string_fuzzy_match_t match) {
+    completions->emplace_back(std::move(comp), std::move(desc), match, flags);
+}
+
+void complete_add(const wchar_t *cmd, bool cmd_is_path, const wcstring &option,
+                  complete_option_type_t option_type, completion_mode_t result_mode,
+                  const wchar_t *condition, const wchar_t *comp, const wchar_t *desc,
+                  complete_flags_t flags) {
+    assert(cmd && "Null command");
+    // option should be empty iff the option type is arguments only.
+    assert(option.empty() == (option_type == option_type_args_only));
+
+    // Lock the lock that allows us to edit the completion entry list.
+    auto completion_set = s_completion_set.acquire();
+    completion_entry_t &c = complete_get_exact_entry(*completion_set, cmd, cmd_is_path);
+
+    // Create our new option.
+    complete_entry_opt_t opt;
+    opt.option = option;
+    opt.type = option_type;
+    opt.result_mode = result_mode;
+
+    if (comp) opt.comp = comp;
+    if (condition) opt.condition = condition;
+    if (desc) opt.desc = desc;
+    opt.flags = flags;
+
+    c.add_option(opt);
+}
+
+void complete_remove(const wcstring &cmd, bool cmd_is_path, const wcstring &option,
+                     complete_option_type_t type) {
+    auto completion_set = s_completion_set.acquire();
+
+    completion_entry_t tmp_entry(cmd, cmd_is_path);
+    auto iter = completion_set->find(tmp_entry);
+    if (iter != completion_set->end()) {
+        // const_cast: See SET_ELEMENTS_ARE_IMMUTABLE.
+        auto &entry = const_cast<completion_entry_t &>(*iter);
+
+        bool delete_it = entry.remove_option(option, type);
+        if (delete_it) {
+            completion_set->erase(iter);
+        }
+    }
+}
+
+void complete_remove_all(const wcstring &cmd, bool cmd_is_path) {
+    auto completion_set = s_completion_set.acquire();
+    completion_entry_t tmp_entry(cmd, cmd_is_path);
+    completion_set->erase(tmp_entry);
+}
+
 completion_list_t complete(const wcstring &cmd_with_subcmds, completion_request_flags_t flags,
                            const operation_context_t &ctx) {
     // Determine the innermost subcommand.
@@ -1802,7 +1821,7 @@ static wcstring completion2string(const complete_entry_opt_t &o, const wcstring
     } else if (o.result_mode.force_files) {
         append_switch(out, L"force-files");
     } else if (o.result_mode.requires_param) {
-        append_switch(out, L"requires-param");
+        append_switch(out, L"require-parameter");
     }
 
     if (is_path)
