@@ -18,7 +18,7 @@
 #include <utility>
 #include <vector>
 
-#include "builtin_bind.h"
+#include "builtins/bind.h"
 #include "common.h"
 #include "env_dispatch.h"
 #include "env_universal_common.h"
@@ -29,6 +29,7 @@
 #include "global_safety.h"
 #include "history.h"
 #include "input.h"
+#include "kill.h"
 #include "path.h"
 #include "proc.h"
 #include "reader.h"
@@ -56,14 +57,20 @@ bool curses_initialized = false;
 /// Does the terminal have the "eat_newline_glitch".
 bool term_has_xn = false;
 
-/// Universal variables global instance. Initialized in env_init.
-static latch_t<env_universal_t> s_universal_variables;
-
 /// Getter for universal variables.
-static env_universal_t *uvars() { return s_universal_variables; }
+/// This is typically initialized in env_init(), and is considered empty before then.
+static acquired_lock<env_universal_t> uvars() {
+    // Leaked to avoid shutdown dtor registration.
+    static auto const s_universal_variables = new owning_lock<env_universal_t>();
+    return s_universal_variables->acquire();
+}
+
+/// Whether we were launched with no_config; in this case setting a uvar instead sets a global.
+static relaxed_atomic_bool_t s_uvar_scope_is_global{false};
 
 bool env_universal_barrier() { return env_stack_t::principal().universal_barrier(); }
 
+namespace {
 struct electric_var_t {
     enum {
         freadonly = 1 << 0,  // May not be modified by the user.
@@ -79,6 +86,7 @@ struct electric_var_t {
 
     bool exports() const { return flags & fexports; }
 
+    static const electric_var_t *for_name(const wchar_t *name);
     static const electric_var_t *for_name(const wcstring &name);
 };
 
@@ -89,6 +97,7 @@ static constexpr const electric_var_t electric_variables[] = {
     {L"SHLVL", electric_var_t::freadonly | electric_var_t::fexports},
     {L"_", electric_var_t::freadonly},
     {L"fish_kill_signal", electric_var_t::freadonly | electric_var_t::fcomputed},
+    {L"fish_killring", electric_var_t::freadonly | electric_var_t::fcomputed},
     {L"fish_pid", electric_var_t::freadonly},
     {L"history", electric_var_t::freadonly | electric_var_t::fcomputed},
     {L"hostname", electric_var_t::freadonly},
@@ -98,32 +107,27 @@ static constexpr const electric_var_t electric_variables[] = {
     {L"umask", electric_var_t::fcomputed},
     {L"version", electric_var_t::freadonly},
 };
-ASSERT_SORT_ORDER(electric_variables, .name);
+ASSERT_SORTED_BY_NAME(electric_variables);
 
-const electric_var_t *electric_var_t::for_name(const wcstring &name) {
-    constexpr auto electric_var_count = sizeof(electric_variables) / sizeof(electric_variables[0]);
-    constexpr auto begin = &electric_variables[0];
-    constexpr auto end = &electric_variables[0] + electric_var_count;
-
-    electric_var_t search{name.c_str(), 0};
-    auto binsearch = std::lower_bound(begin, end, search,
-                                      [&](const electric_var_t &v1, const electric_var_t &v2) {
-                                          return wcscmp(v1.name, v2.name) < 0;
-                                      });
-    if (binsearch != end && wcscmp(name.c_str(), binsearch->name) == 0) {
-        return &*binsearch;
-    }
-    return nullptr;
+const electric_var_t *electric_var_t::for_name(const wchar_t *name) {
+    return get_by_sorted_name(name, electric_variables);
 }
 
+const electric_var_t *electric_var_t::for_name(const wcstring &name) {
+    return electric_var_t::for_name(name.c_str());
+}
+}  // namespace
+
 /// Check if a variable may not be set using the set command.
-static bool is_read_only(const wcstring &key) {
+static bool is_read_only(const wchar_t *key) {
     if (auto ev = electric_var_t::for_name(key)) {
         return ev->readonly();
     }
     return false;
 }
 
+static bool is_read_only(const wcstring &key) { return is_read_only(key.c_str()); }
+
 /// Return true if a variable should become a path variable by default. See #436.
 static bool variable_should_auto_pathvar(const wcstring &name) {
     return string_suffixes_string(L"PATH", name);
@@ -230,7 +234,11 @@ static void setup_path() {
         // _CS_PATH: colon-separated paths to find POSIX utilities
         std::string cspath;
         cspath.resize(confstr(_CS_PATH, nullptr, 0));
-        confstr(_CS_PATH, &cspath[0], cspath.length());
+        if (cspath.length() > 0) {
+            confstr(_CS_PATH, &cspath[0], cspath.length());
+            // remove the trailing null-terminator
+            cspath.resize(cspath.length() - 1);
+        }
 #else
         std::string cspath = "/usr/bin:/bin";  // I doubt this is even necessary
 #endif
@@ -238,13 +246,13 @@ static void setup_path() {
     }
 }
 
-void env_init(const struct config_paths_t *paths /* or NULL */) {
+void env_init(const struct config_paths_t *paths, bool do_uvars, bool default_paths) {
     env_stack_t &vars = env_stack_t::principal();
     // Import environment variables. Walk backwards so that the first one out of any duplicates wins
     // (See issue #2784).
     wcstring key, val;
     const char *const *envp = environ;
-    size_t i = 0;
+    int i = 0;
     while (envp && envp[i]) i++;
     while (i--) {
         const wcstring key_and_val = str2wcstring(envp[i]);  // like foo=bar
@@ -274,6 +282,11 @@ void env_init(const struct config_paths_t *paths /* or NULL */) {
         vars.set_one(FISH_SYSCONFDIR_VAR, ENV_GLOBAL, paths->sysconf);
         vars.set_one(FISH_HELPDIR_VAR, ENV_GLOBAL, paths->doc);
         vars.set_one(FISH_BIN_DIR, ENV_GLOBAL, paths->bin);
+        if (default_paths) {
+            wcstring scstr = paths->data;
+            scstr.append(L"/functions");
+            vars.set_one(L"fish_function_path", ENV_GLOBAL, scstr);
+        }
     }
 
     // Some `su`s keep $USER when changing to root.
@@ -351,16 +364,24 @@ void env_init(const struct config_paths_t *paths /* or NULL */) {
 
     // Set up SHLVL variable. Not we can't use vars.get() because SHLVL is read-only, and therefore
     // was not inherited from the environment.
-    wcstring nshlvl_str = L"1";
-    if (const char *shlvl_var = getenv("SHLVL")) {
-        const wchar_t *end;
-        // TODO: Figure out how to handle invalid numbers better. Shouldn't we issue a diagnostic?
-        long shlvl_i = fish_wcstol(str2wcstring(shlvl_var).c_str(), &end);
-        if (!errno && shlvl_i >= 0) {
-            nshlvl_str = to_string(shlvl_i + 1);
+    if (is_interactive_session()) {
+        wcstring nshlvl_str = L"1";
+        if (const char *shlvl_var = getenv("SHLVL")) {
+            const wchar_t *end;
+            // TODO: Figure out how to handle invalid numbers better. Shouldn't we issue a
+            // diagnostic?
+            long shlvl_i = fish_wcstol(str2wcstring(shlvl_var).c_str(), &end);
+            if (!errno && shlvl_i >= 0) {
+                nshlvl_str = to_string(shlvl_i + 1);
+            }
+        }
+        vars.set_one(L"SHLVL", ENV_GLOBAL | ENV_EXPORT, nshlvl_str);
+    } else {
+        // If we're not interactive, simply pass the value along.
+        if (const char *shlvl_var = getenv("SHLVL")) {
+            vars.set_one(L"SHLVL", ENV_GLOBAL | ENV_EXPORT, str2wcstring(shlvl_var));
         }
     }
-    vars.set_one(L"SHLVL", ENV_GLOBAL | ENV_EXPORT, nshlvl_str);
 
     // initialize the PWD variable if necessary
     // Note we may inherit a virtual PWD that doesn't match what getcwd would return; respect that
@@ -371,7 +392,8 @@ void env_init(const struct config_paths_t *paths /* or NULL */) {
     // (see #7636)
     const char *incoming_pwd_cstr = getenv("PWD");
     wcstring incoming_pwd = incoming_pwd_cstr ? str2wcstring(incoming_pwd_cstr) : wcstring{};
-    if (!incoming_pwd.empty() && incoming_pwd.front() == L'/' &&  paths_are_same_file(incoming_pwd, L".")) {
+    if (!incoming_pwd.empty() && incoming_pwd.front() == L'/' &&
+        paths_are_same_file(incoming_pwd, L".")) {
         vars.set_one(L"PWD", ENV_EXPORT | ENV_GLOBAL, incoming_pwd);
     } else {
         vars.set_pwd_from_getcwd();
@@ -393,24 +415,32 @@ void env_init(const struct config_paths_t *paths /* or NULL */) {
     init_input();
 
     // Complain about invalid config paths.
-    path_emit_config_directory_errors(vars);
+    // HACK: Assume the defaults are correct (in practice this is only --no-config anyway).
+    if (!default_paths) {
+        path_emit_config_directory_messages(vars);
+    }
 
-    // Set up universal variables. The empty string means to use the default path.
-    s_universal_variables.emplace(L"");
-    callback_data_list_t callbacks;
-    s_universal_variables->initialize(callbacks);
-    env_universal_callbacks(&vars, callbacks);
+    // Initialize our uvars if requested.
+    if (!do_uvars) {
+        s_uvar_scope_is_global = true;
+    } else {
+        // Set up universal variables using the default path.
+        callback_data_list_t callbacks;
+        uvars()->initialize(callbacks);
+        env_universal_callbacks(&vars, callbacks);
 
-    // Do not import variables that have the same name and value as
-    // an exported universal variable. See issues #5258 and #5348.
-    for (const auto &kv : uvars()->get_table()) {
-        const wcstring &name = kv.first;
-        const env_var_t &uvar = kv.second;
-        if (!uvar.exports()) continue;
-        // Look for a global exported variable with the same name.
-        maybe_t<env_var_t> global = vars.globals().get(name, ENV_GLOBAL | ENV_EXPORT);
-        if (global && uvar.as_string() == global->as_string()) {
-            vars.globals().remove(name, ENV_GLOBAL | ENV_EXPORT);
+        // Do not import variables that have the same name and value as
+        // an exported universal variable. See issues #5258 and #5348.
+        var_table_t table = uvars()->get_table();
+        for (const auto &kv : table) {
+            const wcstring &name = kv.first;
+            const env_var_t &uvar = kv.second;
+            if (!uvar.exports()) continue;
+            // Look for a global exported variable with the same name.
+            maybe_t<env_var_t> global = vars.globals().get(name, ENV_GLOBAL | ENV_EXPORT);
+            if (global && uvar.as_string() == global->as_string()) {
+                vars.globals().remove(name, ENV_GLOBAL | ENV_EXPORT);
+            }
         }
     }
 }
@@ -432,8 +462,9 @@ struct query_t {
     // Whether any scopes were specified.
     bool has_scope;
 
-    // Whether to search local, global, universal scopes.
+    // Whether to search local, function, global, universal scopes.
     bool local;
+    bool function;
     bool global;
     bool universal;
 
@@ -453,8 +484,9 @@ struct query_t {
     bool user;
 
     explicit query_t(env_mode_flags_t mode) {
-        has_scope = mode & (ENV_LOCAL | ENV_GLOBAL | ENV_UNIVERSAL);
+        has_scope = mode & (ENV_LOCAL | ENV_FUNCTION | ENV_GLOBAL | ENV_UNIVERSAL);
         local = !has_scope || (mode & ENV_LOCAL);
+        function = !has_scope || (mode & ENV_FUNCTION);
         global = !has_scope || (mode & ENV_GLOBAL);
         universal = !has_scope || (mode & ENV_UNIVERSAL);
 
@@ -477,6 +509,14 @@ struct query_t {
             return true;
         }
     }
+
+    bool pathvar_matches(const env_var_t &var) const {
+        if (has_pathvar_unpathvar) {
+            return var.is_pathvar() ? pathvar : unpathvar;
+        } else {
+            return true;
+        }
+    }
 };
 
 // Struct representing one level in the function variable stack.
@@ -510,7 +550,7 @@ class env_node_t {
 }  // namespace
 
 using env_node_ref_t = std::shared_ptr<env_node_t>;
-class env_scoped_impl_t : public environment_t {
+class env_scoped_impl_t : public environment_t, noncopyable_t {
     /// A struct wrapping up parser-local variables. These are conceptually variables that differ in
     /// different fish internal processes.
     struct perproc_data_t {
@@ -534,12 +574,7 @@ class env_scoped_impl_t : public environment_t {
 
     ~env_scoped_impl_t() override = default;
 
-    std::shared_ptr<const null_terminated_array_t<char>> export_array();
-
-    env_scoped_impl_t(env_scoped_impl_t &&) = delete;
-    env_scoped_impl_t(const env_scoped_impl_t &) = delete;
-    void operator=(env_scoped_impl_t &&) = delete;
-    void operator=(const env_scoped_impl_t &) = delete;
+    std::shared_ptr<owning_null_terminated_array_t> export_array();
 
    protected:
     // A linked list of scopes.
@@ -552,7 +587,7 @@ class env_scoped_impl_t : public environment_t {
     perproc_data_t perproc_data_{};
 
     // Exported variable array used by execv.
-    std::shared_ptr<const null_terminated_array_t<char>> export_array_{};
+    std::shared_ptr<owning_null_terminated_array_t> export_array_{};
 
     // Cached list of export generations corresponding to the above export_array_.
     // If this differs from the current export generations then we need to regenerate the array.
@@ -565,6 +600,7 @@ class env_scoped_impl_t : public environment_t {
     // query.
     maybe_t<env_var_t> try_get_computed(const wcstring &key) const;
     maybe_t<env_var_t> try_get_local(const wcstring &key) const;
+    maybe_t<env_var_t> try_get_function(const wcstring &key) const;
     maybe_t<env_var_t> try_get_global(const wcstring &key) const;
     maybe_t<env_var_t> try_get_universal(const wcstring &key) const;
 
@@ -573,7 +609,7 @@ class env_scoped_impl_t : public environment_t {
     void enumerate_generations(const Func &func) const {
         // Our uvars generation count doesn't come from next_export_generation(), so always supply
         // it even if it's 0.
-        func(uvars() ? uvars()->get_export_generation() : 0);
+        func(uvars()->get_export_generation());
         if (globals_->exports()) func(globals_->export_gen);
         for (auto node = locals_; node; node = node->next) {
             if (node->exports()) func(node->export_gen);
@@ -584,7 +620,7 @@ class env_scoped_impl_t : public environment_t {
     bool export_array_needs_regeneration() const;
 
     /// \return a newly allocated export array.
-    std::shared_ptr<const null_terminated_array_t<char>> create_export_array() const;
+    std::shared_ptr<owning_null_terminated_array_t> create_export_array() const;
 };
 
 /// Get the exported variables into a variable table.
@@ -631,26 +667,21 @@ bool env_scoped_impl_t::export_array_needs_regeneration() const {
     return mismatch;
 }
 
-std::shared_ptr<const null_terminated_array_t<char>> env_scoped_impl_t::create_export_array()
-    const {
-    var_table_t table;
-
+std::shared_ptr<owning_null_terminated_array_t> env_scoped_impl_t::create_export_array() const {
     FLOG(env_export, L"create_export_array() recalc");
     var_table_t vals;
     get_exported(this->globals_, vals);
     get_exported(this->locals_, vals);
 
-    if (uvars()) {
-        const wcstring_list_t uni = uvars()->get_names(true, false);
-        for (const wcstring &key : uni) {
-            auto var = uvars()->get(key);
-            assert(var && "Variable should be present in uvars");
-            // Note that std::map::insert does NOT overwrite a value already in the map,
-            // which we depend on here.
-            // Note: Using std::move around make_pair prevents the compiler from implementing
-            // copy elision.
-            vals.insert(std::make_pair(key, std::move(*var)));
-        }
+    const wcstring_list_t uni = uvars()->get_names(true, false);
+    for (const wcstring &key : uni) {
+        auto var = uvars()->get(key);
+        assert(var && "Variable should be present in uvars");
+        // Note that std::map::insert does NOT overwrite a value already in the map,
+        // which we depend on here.
+        // Note: Using std::move around emplace prevents the compiler from implementing
+        // copy elision.
+        vals.emplace(key, std::move(*var));
     }
 
     // Dorky way to add our single exported computed variable.
@@ -665,10 +696,10 @@ std::shared_ptr<const null_terminated_array_t<char>> env_scoped_impl_t::create_e
         str.append(wcs2string(kv.second.as_string()));
         export_list.push_back(std::move(str));
     }
-    return std::make_shared<null_terminated_array_t<char>>(export_list);
+    return std::make_shared<owning_null_terminated_array_t>(std::move(export_list));
 }
 
-std::shared_ptr<const null_terminated_array_t<char>> env_scoped_impl_t::export_array() {
+std::shared_ptr<owning_null_terminated_array_t> env_scoped_impl_t::export_array() {
     ASSERT_IS_NOT_FORKED_CHILD();
     if (export_array_needs_regeneration()) {
         export_array_ = create_export_array();
@@ -695,13 +726,15 @@ maybe_t<env_var_t> env_scoped_impl_t::try_get_computed(const wcstring &key) cons
             return none();
         }
 
-        std::shared_ptr<history_t> history = reader_get_history();
+        std::shared_ptr<history_t> history = commandline_get_state().history;
         if (!history) {
             history = history_t::with_name(history_session_id(*this));
         }
         wcstring_list_t result;
         if (history) history->get_history(result);
         return env_var_t(L"history", std::move(result));
+    } else if (key == L"fish_killring") {
+        return env_var_t(L"fish_killring", kill_entries());
     } else if (key == L"pipestatus") {
         const auto &js = perproc_data().statuses;
         wcstring_list_t result;
@@ -733,32 +766,35 @@ maybe_t<env_var_t> env_scoped_impl_t::try_get_computed(const wcstring &key) cons
 }
 
 maybe_t<env_var_t> env_scoped_impl_t::try_get_local(const wcstring &key) const {
-    auto cursor = locals_;
-    while (cursor) {
-        auto where = cursor->env.find(key);
-        if (where != cursor->env.end()) {
-            return where->second;
-        }
-        cursor = cursor->next;
+    maybe_t<env_var_t> entry;
+    for (auto cur = locals_; cur; cur=cur->next) {
+        if ((entry = cur->find_entry(key))) break;
     }
-    return none();
+    return entry; // this is either the entry or none() from find_entry
+}
+
+maybe_t<env_var_t> env_scoped_impl_t::try_get_function(const wcstring &key) const {
+    maybe_t<env_var_t> entry;
+    auto node = locals_;
+    while (node->next) {
+        node = node->next;
+        // The first node that introduces a new scope is ours.
+        // If this doesn't happen, we go on until we've reached the
+        // topmost local scope.
+        if (node->new_scope) break;
+    }
+    for (auto cur = node; cur; cur=cur->next) {
+        if ((entry = cur->find_entry(key))) break;
+    }
+    return entry; // this is either the entry or none() from find_entry
 }
 
 maybe_t<env_var_t> env_scoped_impl_t::try_get_global(const wcstring &key) const {
-    auto where = globals_->env.find(key);
-    if (where != globals_->env.end()) {
-        return where->second;
-    }
-    return none();
+    return globals_->find_entry(key);
 }
 
 maybe_t<env_var_t> env_scoped_impl_t::try_get_universal(const wcstring &key) const {
-    if (!uvars()) return none();
-    auto var = uvars()->get(key);
-    if (var) {
-        return var;
-    }
-    return none();
+    return uvars()->get(key);
 }
 
 maybe_t<env_var_t> env_scoped_impl_t::get(const wcstring &key, env_mode_flags_t mode) const {
@@ -773,6 +809,9 @@ maybe_t<env_var_t> env_scoped_impl_t::get(const wcstring &key, env_mode_flags_t
     if (!result && query.local) {
         result = try_get_local(key);
     }
+    if (!result && query.function) {
+        result = try_get_function(key);
+    }
     if (!result && query.global) {
         result = try_get_global(key);
     }
@@ -783,6 +822,10 @@ maybe_t<env_var_t> env_scoped_impl_t::get(const wcstring &key, env_mode_flags_t
     if (result && !query.export_matches(*result)) {
         result = none();
     }
+    // Same for pathvars
+    if (result && !query.pathvar_matches(*result)) {
+        result = none();
+    }
     return result;
 }
 
@@ -816,12 +859,12 @@ wcstring_list_t env_scoped_impl_t::get_names(int flags) const {
         }
     }
 
-    if (query.universal && uvars()) {
+    if (query.universal) {
         const wcstring_list_t uni_list = uvars()->get_names(query.exports, query.unexports);
         names.insert(uni_list.begin(), uni_list.end());
     }
 
-    return wcstring_list_t(names.begin(), names.end());
+    return {names.begin(), names.end()};
 }
 
 /// Recursive helper to snapshot a series of nodes.
@@ -846,6 +889,7 @@ std::shared_ptr<environment_t> env_scoped_impl_t::snapshot() const {
 }
 
 // A struct that wraps up the result of setting or removing a variable.
+namespace {
 struct mod_result_t {
     // The publicly visible status of the set call.
     int status{ENV_OK};
@@ -858,6 +902,7 @@ struct mod_result_t {
 
     explicit mod_result_t(int status) : status(status) {}
 };
+}  // namespace
 
 /// A mutable subclass of env_scoped_impl_t.
 class env_stack_impl_t final : public env_scoped_impl_t {
@@ -1023,10 +1068,8 @@ void env_stack_impl_t::set_in_node(const env_node_ref_t &node, const wcstring &k
         val = colon_split(val);
     }
 
-    var = var.setting_vals(std::move(val))
-              .setting_exports(res_exports)
-              .setting_pathvar(res_pathvar)
-              .setting_read_only(is_read_only(key));
+    var =
+        var.setting_vals(std::move(val)).setting_exports(res_exports).setting_pathvar(res_pathvar);
 
     // Perhaps mark that this node contains an exported variable, or shadows an exported variable.
     // If so regenerate the export list.
@@ -1085,7 +1128,6 @@ maybe_t<int> env_stack_impl_t::try_set_electric(const wcstring &key, const query
 void env_stack_impl_t::set_universal(const wcstring &key, wcstring_list_t val,
                                      const query_t &query) {
     ASSERT_IS_MAIN_THREAD();
-    if (!uvars()) return;
     auto oldvar = uvars()->get(key);
     // Resolve whether or not to export.
     bool exports = false;
@@ -1152,15 +1194,33 @@ mod_result_t env_stack_impl_t::set(const wcstring &key, env_mode_flags_t mode,
     mod_result_t result{ENV_OK};
     if (query.has_scope) {
         // The user requested a particular scope.
-        if (query.universal) {
+        //
+        // If we don't have uvars, fall back to using globals
+        if (query.universal && !s_uvar_scope_is_global) {
             set_universal(key, std::move(val), query);
             result.uvar_modified = true;
-        } else if (query.global) {
+        } else if (query.global || (query.universal && s_uvar_scope_is_global)) {
             set_in_node(globals_, key, std::move(val), flags);
             result.global_modified = true;
         } else if (query.local) {
             assert(locals_ != globals_ && "Locals should not be globals");
             set_in_node(locals_, key, std::move(val), flags);
+        } else if (query.function) {
+            // "Function" scope is:
+            // Either the topmost local scope of the nearest function,
+            // or the top-level local scope if no function exists.
+            //
+            // This is distinct from the unspecified scope,
+            // which is the global scope if no function exists.
+            auto node = locals_;
+            while (node->next) {
+                node = node->next;
+                // The first node that introduces a new scope is ours.
+                // If this doesn't happen, we go on until we've reached the
+                // topmost local scope.
+                if (node->new_scope) break;
+            }
+            set_in_node(node, key, std::move(val), flags);
         } else {
             DIE("Unknown scope");
         }
@@ -1171,13 +1231,13 @@ mod_result_t env_stack_impl_t::set(const wcstring &key, env_mode_flags_t mode,
         // Existing global variable.
         set_in_node(node, key, std::move(val), flags);
         result.global_modified = true;
-    } else if (uvars() && uvars()->get(key)) {
+    } else if (uvars()->get(key)) {
         // Existing universal variable.
         set_universal(key, std::move(val), query);
         result.uvar_modified = true;
     } else {
         // Unspecified scope with no existing variables.
-        auto node = resolve_unspecified_scope();
+        node = resolve_unspecified_scope();
         assert(node && "Should always resolve some scope");
         set_in_node(node, key, std::move(val), flags);
         result.global_modified = (node == globals_);
@@ -1193,20 +1253,24 @@ mod_result_t env_stack_impl_t::remove(const wcstring &key, int mode) {
         return mod_result_t{ENV_SCOPE};
     }
 
-    // Helper to remove from uvars.
-    auto remove_from_uvars = [&] { return uvars() && uvars()->remove(key); };
-
     mod_result_t result{ENV_OK};
     if (query.has_scope) {
         // The user requested erasing from a particular scope.
         if (query.universal) {
-            result.status = remove_from_uvars() ? ENV_OK : ENV_NOT_FOUND;
+            result.status = uvars()->remove(key) ? ENV_OK : ENV_NOT_FOUND;
             result.uvar_modified = true;
         } else if (query.global) {
             result.status = remove_from_chain(globals_, key) ? ENV_OK : ENV_NOT_FOUND;
             result.global_modified = true;
         } else if (query.local) {
             result.status = remove_from_chain(locals_, key) ? ENV_OK : ENV_NOT_FOUND;
+        } else if (query.function) {
+            auto node = locals_;
+            while (node->next) {
+                node = node->next;
+                if (node->new_scope) break;
+            }
+            result.status = remove_from_chain(node, key) ? ENV_OK : ENV_NOT_FOUND;
         } else {
             DIE("Unknown scope");
         }
@@ -1214,7 +1278,7 @@ mod_result_t env_stack_impl_t::remove(const wcstring &key, int mode) {
         // pass
     } else if (remove_from_chain(globals_, key)) {
         result.global_modified = true;
-    } else if (remove_from_uvars()) {
+    } else if (uvars()->remove(key)) {
         result.uvar_modified = true;
     } else {
         result.status = ENV_NOT_FOUND;
@@ -1229,7 +1293,7 @@ bool env_stack_t::universal_barrier() {
     if (!is_principal()) return false;
 
     ASSERT_IS_MAIN_THREAD();
-    if (!uvars()) return false;
+    if (s_uvar_scope_is_global) return false;
 
     callback_data_list_t callbacks;
     bool changed = uvars()->sync(callbacks);
@@ -1278,8 +1342,7 @@ maybe_t<env_var_t> env_stack_t::get(const wcstring &key, env_mode_flags_t mode)
 
 wcstring_list_t env_stack_t::get_names(int flags) const { return acquire_impl()->get_names(flags); }
 
-int env_stack_t::set(const wcstring &key, env_mode_flags_t mode, wcstring_list_t vals,
-                     std::vector<event_t> *out_events) {
+int env_stack_t::set(const wcstring &key, env_mode_flags_t mode, wcstring_list_t vals) {
     // Historical behavior.
     if (vals.size() == 1 && (key == L"PWD" || key == L"HOME")) {
         path_make_canonical(vals.front());
@@ -1301,9 +1364,6 @@ int env_stack_t::set(const wcstring &key, env_mode_flags_t mode, wcstring_list_t
         if (ret.global_modified || is_principal()) {
             env_dispatch_var_change(key, *this);
         }
-        if (out_events) {
-            out_events->push_back(event_t::variable(key, {L"VARIABLE", L"SET", key}));
-        }
     }
     // If the principal stack modified universal variables, then post a barrier.
     if (ret.uvar_modified && is_principal()) {
@@ -1312,28 +1372,23 @@ int env_stack_t::set(const wcstring &key, env_mode_flags_t mode, wcstring_list_t
     return ret.status;
 }
 
-int env_stack_t::set_one(const wcstring &key, env_mode_flags_t mode, wcstring val,
-                         std::vector<event_t> *out_events) {
+int env_stack_t::set_one(const wcstring &key, env_mode_flags_t mode, wcstring val) {
     wcstring_list_t vals;
     vals.push_back(std::move(val));
-    return set(key, mode, std::move(vals), out_events);
+    return set(key, mode, std::move(vals));
 }
 
-int env_stack_t::set_empty(const wcstring &key, env_mode_flags_t mode,
-                           std::vector<event_t> *out_events) {
-    return set(key, mode, {}, out_events);
+int env_stack_t::set_empty(const wcstring &key, env_mode_flags_t mode) {
+    return set(key, mode, {});
 }
 
-int env_stack_t::remove(const wcstring &key, int mode, std::vector<event_t> *out_events) {
+int env_stack_t::remove(const wcstring &key, int mode) {
     mod_result_t ret = acquire_impl()->remove(key, mode);
     if (ret.status == ENV_OK) {
         if (ret.global_modified || is_principal()) {
             // Important to not hold the lock here.
             env_dispatch_var_change(key, *this);
         }
-        if (out_events) {
-            out_events->push_back(event_t::variable(key, {L"VARIABLE", L"ERASE", key}));
-        }
     }
     if (ret.uvar_modified && is_principal()) {
         universal_barrier();
@@ -1341,7 +1396,7 @@ int env_stack_t::remove(const wcstring &key, int mode, std::vector<event_t> *out
     return ret.status;
 }
 
-std::shared_ptr<const null_terminated_array_t<char>> env_stack_t::export_arr() {
+std::shared_ptr<owning_null_terminated_array_t> env_stack_t::export_arr() {
     return acquire_impl()->export_array();
 }
 
