@@ -53,22 +53,28 @@
 #include "reader.h"
 #include "screen.h"
 #include "termsize.h"
+#include "trace.h"
 #include "wutil.h"  // IWYU pragma: keep
 
 #define DEFAULT_TERM1 "ansi"
 #define DEFAULT_TERM2 "dumb"
 
 /// List of all locale environment variable names that might trigger (re)initializing the locale
-/// subsystem.
-static const wcstring_list_t locale_variables({L"LANG", L"LANGUAGE", L"LC_ALL", L"LC_ADDRESS",
-                                               L"LC_COLLATE", L"LC_CTYPE", L"LC_IDENTIFICATION",
-                                               L"LC_MEASUREMENT", L"LC_MESSAGES", L"LC_MONETARY",
-                                               L"LC_NAME", L"LC_NUMERIC", L"LC_PAPER",
-                                               L"LC_TELEPHONE", L"LC_TIME", L"LOCPATH"});
+/// subsystem. These are only the variables we're possibly interested in.
+static const wcstring locale_variables[] = {L"LANG",
+                                            L"LANGUAGE",
+                                            L"LC_ALL",
+                                            L"LC_COLLATE",
+                                            L"LC_CTYPE",
+                                            L"LC_MESSAGES",
+                                            L"LC_NUMERIC",
+                                            L"LC_TIME",
+                                            L"fish_allow_singlebyte_locale",
+                                            L"LOCPATH"};
 
 /// List of all curses environment variable names that might trigger (re)initializing the curses
 /// subsystem.
-static const wcstring_list_t curses_variables({L"TERM", L"TERMINFO", L"TERMINFO_DIRS"});
+static const wcstring curses_variables[] = {L"TERM", L"TERMINFO", L"TERMINFO_DIRS"};
 
 class var_dispatch_table_t {
     using named_callback_t = std::function<void(const wcstring &, env_stack_t &)>;
@@ -139,7 +145,7 @@ static void handle_timezone(const wchar_t *env_var_name, const environment_t &va
     const auto var = vars.get(env_var_name, ENV_DEFAULT);
     FLOGF(env_dispatch, L"handle_timezone() current timezone var: |%ls| => |%ls|", env_var_name,
           !var ? L"MISSING" : var->as_string().c_str());
-    const std::string &name = wcs2string(env_var_name);
+    std::string name = wcs2string(env_var_name);
     if (var.missing_or_empty()) {
         unsetenv_lock(name.c_str());
     } else {
@@ -149,11 +155,11 @@ static void handle_timezone(const wchar_t *env_var_name, const environment_t &va
     tzset();
 }
 
-/// Update the value of g_guessed_fish_emoji_width
+/// Update the value of g_fish_emoji_width
 static void guess_emoji_width(const environment_t &vars) {
     if (auto width_str = vars.get(L"fish_emoji_width")) {
         int new_width = fish_wcstol(width_str->as_string().c_str());
-        g_fish_emoji_width = std::max(0, new_width);
+        g_fish_emoji_width = std::min(2, std::max(1, new_width));
         FLOGF(term_support, "'fish_emoji_width' preference: %d, overwriting default",
               g_fish_emoji_width);
         return;
@@ -172,19 +178,18 @@ static void guess_emoji_width(const environment_t &vars) {
 
     if (term == L"Apple_Terminal" && version >= 400) {
         // Apple Terminal on High Sierra
-        g_guessed_fish_emoji_width = 2;
+        g_fish_emoji_width = 2;
         FLOGF(term_support, "default emoji width: 2 for %ls", term.c_str());
     } else if (term == L"iTerm.app") {
-        // iTerm2 defaults to Unicode 8 sizes.
-        // See https://gitlab.com/gnachman/iterm2/wikis/unicodeversionswitching
-        g_guessed_fish_emoji_width = 1;
-        FLOGF(term_support, "default emoji width: 1");
+        // iTerm2 now defaults to Unicode 9 sizes for anything after macOS 10.12.
+        g_fish_emoji_width = 2;
+        FLOGF(term_support, "default emoji width for iTerm: 2");
     } else {
         // Default to whatever system wcwidth says to U+1F603,
-        // but only if it's at least 1.
+        // but only if it's at least 1 and at most 2.
         int w = wcwidth(L'😃');
-        g_guessed_fish_emoji_width = w > 0 ? w : 1;
-        FLOGF(term_support, "default emoji width: %d", g_guessed_fish_emoji_width);
+        g_fish_emoji_width = std::min(2, std::max(1, w));
+        FLOGF(term_support, "default emoji width: %d", g_fish_emoji_width);
     }
 }
 
@@ -197,20 +202,16 @@ void env_dispatch_var_change(const wcstring &key, env_stack_t &vars) {
     s_var_dispatch_table->dispatch(key, vars);
 }
 
-/// Universal variable callback function. This function makes sure the proper events are triggered
-/// when an event occurs.
-static void universal_callback(env_stack_t *stack, const callback_data_t &cb) {
-    const wchar_t *op = cb.is_erase() ? L"ERASE" : L"SET";
-
-    env_dispatch_var_change(cb.key, *stack);
-
-    // TODO: eliminate this principal_parser. Need to rationalize how multiple threads work here.
-    event_fire(parser_t::principal_parser(), event_t::variable(cb.key, {L"VARIABLE", op, cb.key}));
-}
-
+// Trigger events due to a universal variable changing.
 void env_universal_callbacks(env_stack_t *stack, const callback_data_list_t &callbacks) {
     for (const callback_data_t &cb : callbacks) {
-        universal_callback(stack, cb);
+        env_dispatch_var_change(cb.key, *stack);
+
+        // TODO: eliminate this principal_parser. Need to rationalize how multiple threads work
+        // here.
+        event_t evt =
+            cb.is_erase() ? event_t::variable_erase(cb.key) : event_t::variable_set(cb.key);
+        event_fire(parser_t::principal_parser(), evt);
     }
 }
 
@@ -235,6 +236,10 @@ static void handle_fish_history_change(const env_stack_t &vars) {
     reader_change_history(history_session_id(vars));
 }
 
+void handle_autosuggestion_change(const env_stack_t &vars) {
+    reader_set_autosuggestion_enabled(vars);
+}
+
 static void handle_function_path_change(const env_stack_t &vars) {
     UNUSED(vars);
     function_invalidate_path();
@@ -260,11 +265,42 @@ static void handle_curses_change(const environment_t &vars) {
     init_curses(vars);
 }
 
+/// Whether to use posix_spawn when possible.
+static relaxed_atomic_bool_t g_use_posix_spawn{false};
+
+bool get_use_posix_spawn() { return g_use_posix_spawn; }
+
+extern "C" {
+const char *gnu_get_libc_version();
+}
+
+static bool allow_use_posix_spawn() {
+    // OpenBSD's posix_spawn returns status 127, instead of erroring with ENOEXEC, when faced with a
+    // shebangless script. Disable posix_spawn on OpenBSD.
+#if defined(__OpenBSD__)
+    return false;
+#endif
+    bool result = true;
+    // uClibc defines __GLIBC__.
+#if defined(__GLIBC__) && !defined(__UCLIBC__)
+    // Disallow posix_spawn entirely on glibc < 2.24.
+    // See #8021.
+    if (!__GLIBC_PREREQ(2, 24)) {
+        result = false;
+    }
+#endif
+    return result;
+}
+
 static void handle_fish_use_posix_spawn_change(const environment_t &vars) {
-    // note this defaults to true
-    auto use_posix_spawn = vars.get(L"fish_use_posix_spawn");
-    g_use_posix_spawn =
-        use_posix_spawn.missing_or_empty() ? true : bool_from_string(use_posix_spawn->as_string());
+    // Note if the variable is missing or empty, we default to true if allowed.
+    if (!allow_use_posix_spawn()) {
+        g_use_posix_spawn = false;
+    } else if (auto var = vars.get(L"fish_use_posix_spawn")) {
+        g_use_posix_spawn = var->empty() || bool_from_string(var->as_string());
+    } else {
+        g_use_posix_spawn = true;
+    }
 }
 
 /// Allow the user to override the limit on how much data the `read` command will process.
@@ -281,6 +317,10 @@ static void handle_read_limit_change(const environment_t &vars) {
     }
 }
 
+static void handle_fish_trace(const environment_t &vars) {
+    trace_set_enabled(!vars.get(L"fish_trace").missing_or_empty());
+}
+
 /// Populate the dispatch table used by `env_dispatch_var_change()` to efficiently call the
 /// appropriate function to handle a change to a variable.
 /// Note this returns a new-allocated value that we expect to leak.
@@ -305,8 +345,10 @@ static std::unique_ptr<const var_dispatch_table_t> create_dispatch_table() {
     var_dispatch_table->add(L"fish_function_path", handle_function_path_change);
     var_dispatch_table->add(L"fish_read_limit", handle_read_limit_change);
     var_dispatch_table->add(L"fish_history", handle_fish_history_change);
+    var_dispatch_table->add(L"fish_autosuggestion_enabled", handle_autosuggestion_change);
     var_dispatch_table->add(L"TZ", handle_tz_change);
     var_dispatch_table->add(L"fish_use_posix_spawn", handle_fish_use_posix_spawn_change);
+    var_dispatch_table->add(L"fish_trace", handle_fish_trace);
 
     // This std::move is required to avoid a build error on old versions of libc++ (#5801),
     // but it causes a different warning under newer versions of GCC (observed under GCC 9.3.0,
@@ -329,6 +371,7 @@ static void run_inits(const environment_t &vars) {
     update_wait_on_escape_ms(vars);
     handle_read_limit_change(vars);
     handle_fish_use_posix_spawn_change(vars);
+    handle_fish_trace(vars);
 }
 
 /// Updates our idea of whether we support term256 and term24bit (see issue #10222).
@@ -350,21 +393,9 @@ static void update_fish_color_support(const environment_t &vars) {
         support_term256 = true;
         FLOGF(term_support, L"256 color support enabled for TERM=%ls", term.c_str());
     } else if (term.find(L"xterm") != wcstring::npos) {
-        // Assume that all 'xterm's can handle 256, except for Terminal.app from Snow Leopard
-        wcstring term_program;
-        if (auto tp = vars.get(L"TERM_PROGRAM")) term_program = tp->as_string();
-        if (term_program == L"Apple_Terminal") {
-            auto tpv = vars.get(L"TERM_PROGRAM_VERSION");
-            if (tpv && fish_wcstod(tpv->as_string().c_str(), nullptr) > 299) {
-                // OS X Lion is version 299+, it has 256 color support (see github Wiki)
-                support_term256 = true;
-                FLOGF(term_support, L"256 color support enabled for TERM=%ls on Terminal.app",
-                      term.c_str());
-            }
-        } else {
-            support_term256 = true;
-            FLOGF(term_support, L"256 color support enabled for TERM=%ls", term.c_str());
-        }
+        // Assume that all 'xterm's can handle 25
+        support_term256 = true;
+        FLOGF(term_support, L"256 color support enabled for TERM=%ls", term.c_str());
     } else if (cur_term != nullptr) {
         // See if terminfo happens to identify 256 colors
         support_term256 = (max_colors >= 256);
@@ -461,8 +492,8 @@ static bool initialize_curses_using_fallback(const char *term) {
 /// One situation in which this breaks down is with screen, since screen supports setting the
 /// terminal title if the underlying terminal does so, but will print garbage on terminals that
 /// don't. Since we can't see the underlying terminal below screen there is no way to fix this.
-static const wchar_t *const title_terms[] = {L"xterm",  L"screen", L"tmux",
-                                             L"nxterm", L"rxvt",   L"alacritty"};
+static const wchar_t *const title_terms[] = {L"xterm", L"screen",    L"tmux",   L"nxterm",
+                                             L"rxvt",  L"alacritty", L"wezterm"};
 static bool does_term_support_setting_title(const environment_t &vars) {
     const auto term_var = vars.get(L"TERM");
     if (term_var.missing_or_empty()) return false;
@@ -531,6 +562,10 @@ static void init_curses(const environment_t &vars) {
     curses_initialized = true;
 }
 
+static constexpr const char *utf8_locales[] = {
+    "C.UTF-8", "en_US.UTF-8", "en_GB.UTF-8", "de_DE.UTF-8", "C.utf8", "UTF-8",
+};
+
 /// Initialize the locale subsystem.
 static void init_locale(const environment_t &vars) {
     // We have to make a copy because the subsequent setlocale() call to change the locale will
@@ -539,7 +574,7 @@ static void init_locale(const environment_t &vars) {
 
     for (const auto &var_name : locale_variables) {
         const auto var = vars.get(var_name, ENV_EXPORT);
-        const std::string &name = wcs2string(var_name);
+        std::string name = wcs2string(var_name);
         if (var.missing_or_empty()) {
             FLOGF(env_locale, L"locale var %s missing or empty", name.c_str());
             unsetenv_lock(name.c_str());
@@ -551,6 +586,34 @@ static void init_locale(const environment_t &vars) {
     }
 
     char *locale = setlocale(LC_ALL, "");
+
+    // Try to get a multibyte-capable encoding
+    // A "C" locale is broken for our purposes - any wchar functions will break on it.
+    // So we try *really really really hard* to not have one.
+    bool fix_locale = true;
+    if (auto allow_c = vars.get(L"fish_allow_singlebyte_locale")) {
+        fix_locale = allow_c.missing_or_empty() ? true : !bool_from_string(allow_c->as_string());
+    }
+    if (fix_locale && MB_CUR_MAX == 1) {
+        FLOGF(env_locale, L"Have singlebyte locale, trying to fix");
+        for (auto loc : utf8_locales) {
+            setlocale(LC_CTYPE, loc);
+            if (MB_CUR_MAX > 1) {
+                FLOGF(env_locale, L"Fixed locale: '%s'", loc);
+                break;
+            }
+        }
+        if (MB_CUR_MAX == 1) {
+            FLOGF(env_locale, L"Failed to fix locale");
+        }
+    }
+    // We *always* use a C-locale for numbers,
+    // because we always want "." except for in printf.
+    setlocale(LC_NUMERIC, "C");
+
+    // See that we regenerate our special locale for numbers.
+    fish_invalidate_numeric_locale();
+
     fish_setlocale();
     FLOGF(env_locale, L"init_locale() setlocale(): '%s'", locale);
 
@@ -570,9 +633,6 @@ static void init_locale(const environment_t &vars) {
 /// Returns true if we think the terminal supports setting its title.
 bool term_supports_setting_title() { return can_set_term_title; }
 
-/// Miscellaneous variables.
-bool g_use_posix_spawn = false;
-
 // Limit `read` to 100 MiB (bytes not wide chars) by default. This can be overridden by the
 // fish_read_limit variable.
 size_t read_byte_limit = 100 * 1024 * 1024;
