@@ -25,6 +25,7 @@
 #include "wcstringutil.h"
 #include "wutil.h"  // IWYU pragma: keep
 
+namespace {
 class pending_signals_t {
     static constexpr size_t SIGNAL_COUNT = NSIG;
 
@@ -42,8 +43,8 @@ class pending_signals_t {
     pending_signals_t() = default;
 
     /// No copying.
-    pending_signals_t(const pending_signals_t &);
-    void operator=(const pending_signals_t &);
+    pending_signals_t(const pending_signals_t &) = delete;
+    pending_signals_t &operator=(const pending_signals_t &) = delete;
 
     /// Mark a signal as pending. This may be called from a signal handler.
     /// We expect only one signal handler to execute at once.
@@ -83,6 +84,7 @@ class pending_signals_t {
         return result;
     }
 };
+}  // namespace
 
 static pending_signals_t s_pending_signals;
 
@@ -141,7 +143,7 @@ static bool handler_matches(const event_handler_t &classv, const event_t &instan
 }
 
 /// Test if specified event is blocked.
-static int event_is_blocked(parser_t &parser, const event_t &e) {
+static bool event_is_blocked(parser_t &parser, const event_t &e) {
     (void)e;
     const block_t *block;
     size_t idx = 0;
@@ -222,6 +224,15 @@ event_handler_list_t event_get_function_handlers(const wcstring &name) {
     return result;
 }
 
+enum_set_t<event_type_t> event_get_handled_types() {
+    enum_set_t<event_type_t> result{};
+    auto handlers = s_event_handlers.acquire();
+    for (const shared_ptr<event_handler_t> &eh : *handlers) {
+        result.set(eh->desc.type);
+    }
+    return result;
+}
+
 bool event_is_signal_observed(int sig) {
     // We are in a signal handler! Don't allocate memory, etc.
     bool result = false;
@@ -250,7 +261,8 @@ static void event_fire_internal(parser_t &parser, const event_t &event) {
     };
     std::vector<firing_handler_t> fire;
     {
-        for (const auto &handler : *s_event_handlers.acquire()) {
+        auto event_handlers = s_event_handlers.acquire();
+        for (const auto &handler : *event_handlers) {
             // Check if this event is a match.
             bool only_once = false;
             if (!handler_matches(*handler, event, only_once)) {
@@ -428,7 +440,7 @@ void event_print(io_streams_t &streams, const wcstring &type_filter) {
                   }
                   switch (d1.type) {
                       case event_type_t::signal:
-                          return d1.signal < d2.signal;
+                          return d1.param1.signal < d2.param1.signal;
                       case event_type_t::process_exit:
                           return d1.param1.pid < d2.param1.pid;
                       case event_type_t::job_exit:
@@ -508,10 +520,18 @@ event_description_t event_description_t::generic(wcstring str) {
 }
 
 // static
-event_t event_t::variable(wcstring name, wcstring_list_t args) {
+event_t event_t::variable_erase(wcstring name) {
     event_t evt{event_type_t::variable};
+    evt.arguments = {L"VARIABLE", L"ERASE", name};
+    evt.desc.str_param1 = std::move(name);
+    return evt;
+}
+
+// static
+event_t event_t::variable_set(wcstring name) {
+    event_t evt{event_type_t::variable};
+    evt.arguments = {L"VARIABLE", L"SET", name};
     evt.desc.str_param1 = std::move(name);
-    evt.arguments = std::move(args);
     return evt;
 }
 
@@ -527,20 +547,20 @@ event_t event_t::process_exit(pid_t pid, int status) {
 }
 
 // static
-event_t event_t::job_exit(pid_t pid, internal_job_id_t jid) {
+event_t event_t::job_exit(pid_t pgid, internal_job_id_t jid) {
     event_t evt{event_type_t::job_exit};
-    evt.desc.param1.jobspec = {pid, jid};
+    evt.desc.param1.jobspec = {pgid, jid};
     evt.arguments.reserve(3);
     evt.arguments.push_back(L"JOB_EXIT");
-    evt.arguments.push_back(to_string(pid));
+    evt.arguments.push_back(to_string(pgid));
     evt.arguments.push_back(L"0");  // historical
     return evt;
 }
 
 // static
-event_t event_t::caller_exit(uint64_t caller_id, int job_id) {
+event_t event_t::caller_exit(uint64_t internal_job_id, int job_id) {
     event_t evt{event_type_t::caller_exit};
-    evt.desc.param1.caller_id = caller_id;
+    evt.desc.param1.caller_id = internal_job_id;
     evt.arguments.reserve(3);
     evt.arguments.push_back(L"JOB_EXIT");
     evt.arguments.push_back(to_string(job_id));
