@@ -2,6 +2,7 @@
 #include "config.h"
 
 #include <errno.h>
+#include <signal.h>
 #include <unistd.h>
 
 #include <cstring>
@@ -10,6 +11,7 @@
 #endif
 #include <stdio.h>
 #include <stdlib.h>
+#include <sys/select.h>
 #include <sys/time.h>
 #include <sys/types.h>
 
@@ -57,10 +59,10 @@ enum {
 };
 using readb_result_t = int;
 
-static readb_result_t readb(int in_fd, bool queue_is_empty) {
+static readb_result_t readb(int in_fd) {
     assert(in_fd >= 0 && "Invalid in fd");
     universal_notifier_t& notifier = universal_notifier_t::default_notifier();
-    select_wrapper_t fdset;
+    fd_readable_set_t fdset;
     for (;;) {
         fdset.clear();
         fdset.add(in_fd);
@@ -75,13 +77,13 @@ static readb_result_t readb(int in_fd, bool queue_is_empty) {
 
         // Get its suggested delay (possibly none).
         // Note a 0 here means do not poll.
-        uint64_t timeout = select_wrapper_t::kNoTimeout;
+        uint64_t timeout = fd_readable_set_t::kNoTimeout;
         if (uint64_t usecs_delay = notifier.usec_delay_between_polls()) {
             timeout = usecs_delay;
         }
 
         // Here's where we call select().
-        int select_res = fdset.select(timeout);
+        int select_res = fdset.check_readable(timeout);
         if (select_res < 0) {
             if (errno == EINTR || errno == EAGAIN) {
                 // A signal.
@@ -98,9 +100,7 @@ static readb_result_t readb(int in_fd, bool queue_is_empty) {
         // This may come about through readability, or through a call to poll().
         if ((fdset.test(notifier_fd) && notifier.notification_fd_became_readable(notifier_fd)) ||
             notifier.poll()) {
-            if (env_universal_barrier() && !queue_is_empty) {
-                return readb_uvar_notified;
-            }
+            return readb_uvar_notified;
         }
 
         // Check stdin.
@@ -169,7 +169,7 @@ char_event_t input_event_queue_t::readch() {
             return mevt.acquire();
         }
 
-        readb_result_t rr = readb(in_, queue_.empty());
+        readb_result_t rr = readb(in_);
         switch (rr) {
             case readb_eof:
                 return char_event_type_t::eof;
@@ -180,6 +180,7 @@ char_event_t input_event_queue_t::readch() {
                 break;
 
             case readb_uvar_notified:
+                env_universal_barrier();
                 break;
 
             case readb_ioport_notified:
@@ -224,9 +225,27 @@ maybe_t<char_event_t> input_event_queue_t::readch_timed() {
     if (auto evt = try_pop()) {
         return evt;
     }
-    const uint64_t usec_per_msec = 1000;
-    uint64_t timeout_usec = static_cast<uint64_t>(wait_on_escape_ms) * usec_per_msec;
-    if (select_wrapper_t::is_fd_readable(in_, timeout_usec)) {
+    // We are not prepared to handle a signal immediately; we only want to know if we get input on
+    // our fd before the timeout. Use pselect to block all signals; we will handle signals
+    // before the next call to getch().
+    sigset_t sigs;
+    sigfillset(&sigs);
+
+    // pselect expects timeouts in nanoseconds.
+    const uint64_t nsec_per_msec = 1000 * 1000;
+    const uint64_t nsec_per_sec = nsec_per_msec * 1000;
+    const uint64_t wait_nsec = wait_on_escape_ms * nsec_per_msec;
+    struct timespec timeout;
+    timeout.tv_sec = (wait_nsec) / nsec_per_sec;
+    timeout.tv_nsec = (wait_nsec) % nsec_per_sec;
+
+    // We have one fd of interest.
+    fd_set fdset;
+    FD_ZERO(&fdset);
+    FD_SET(in_, &fdset);
+
+    int res = pselect(in_ + 1, &fdset, nullptr, nullptr, &timeout, &sigs);
+    if (res > 0) {
         return readch();
     }
     return none();
@@ -236,6 +255,15 @@ void input_event_queue_t::push_back(const char_event_t& ch) { queue_.push_back(c
 
 void input_event_queue_t::push_front(const char_event_t& ch) { queue_.push_front(ch); }
 
+void input_event_queue_t::promote_interruptions_to_front() {
+    // Find the first sequence of non-char events.
+    // EOF is considered a char: we don't want to pull EOF in front of real chars.
+    auto is_char = [](const char_event_t& ch) { return ch.is_char() || ch.is_eof(); };
+    auto first = std::find_if_not(queue_.begin(), queue_.end(), is_char);
+    auto last = std::find_if(first, queue_.end(), is_char);
+    std::rotate(queue_.begin(), first, last);
+}
+
 void input_event_queue_t::prepare_to_select() {}
 void input_event_queue_t::select_interrupted() {}
 input_event_queue_t::~input_event_queue_t() = default;
