@@ -6,7 +6,12 @@
 #include "path.h"
 
 #include <errno.h>
+#include <sys/mount.h>
+#include <sys/param.h>
 #include <sys/stat.h>
+#if defined(__linux__)
+#include <sys/statfs.h>
+#endif
 #include <unistd.h>
 
 #include <cstring>
@@ -33,6 +38,10 @@ const wcstring_list_t dflt_pathsv({L"/bin", L"/usr/bin", PREFIX L"/bin"});
 
 static bool path_get_path_core(const wcstring &cmd, wcstring *out_path,
                                const maybe_t<env_var_t> &bin_path_var) {
+    // Unix paths can't include a NULL-byte, that's the separator.
+    // If we let this through, we'd end up checking up to the NULL,
+    // so we'd get the wrong path.
+    if (cmd.find(L'\0') != wcstring::npos) return false;
     // If the command has a slash, it must be an absolute or relative path and thus we don't bother
     // looking for a matching command.
     if (cmd.find(L'/') != wcstring::npos) {
@@ -89,15 +98,50 @@ bool path_get_path(const wcstring &cmd, wcstring *out_path, const environment_t
     return path_get_path_core(cmd, out_path, vars.get(L"PATH"));
 }
 
-bool path_is_executable(const std::string &path) {
+static bool path_is_executable(const std::string &path) {
     if (access(path.c_str(), X_OK)) return false;
     struct stat buff;
     if (stat(path.c_str(), &buff) == -1) {
         if (errno != EACCES) wperror(L" stat");
         return false;
     }
-    if (!S_ISREG(buff.st_mode)) return false;
-    return true;
+    return S_ISREG(buff.st_mode);
+}
+
+/// \return whether the given path is on a remote filesystem.
+static dir_remoteness_t path_remoteness(const wcstring &path) {
+    std::string narrow = wcs2string(path);
+#if defined(__linux__)
+    struct statfs buf {};
+    if (statfs(narrow.c_str(), &buf) < 0) {
+        return dir_remoteness_t::unknown;
+    }
+    // Linux has constants for these like NFS_SUPER_MAGIC, SMB_SUPER_MAGIC, CIFS_MAGIC_NUMBER but
+    // these are in varying headers. Simply hard code them.
+    // NOTE: The cast is necessary for 32-bit systems because of the 4-byte CIFS_MAGIC_NUMBER
+    switch (static_cast<unsigned int>(buf.f_type)) {
+        case 0x6969:       // NFS_SUPER_MAGIC
+        case 0x517B:       // SMB_SUPER_MAGIC
+        case 0xFE534D42U:  // SMB2_MAGIC_NUMBER - not in the manpage
+        case 0xFF534D42U:  // CIFS_MAGIC_NUMBER
+            return dir_remoteness_t::remote;
+        default:
+            // Other FSes are assumed local.
+            return dir_remoteness_t::local;
+    }
+#elif defined(ST_LOCAL)
+    // ST_LOCAL is a flag to statvfs, which is itself standardized.
+    // In practice the only system to use this path is NetBSD.
+    struct statvfs buf {};
+    if (statvfs(narrow.c_str(), &buf) < 0) return dir_remoteness_t::unknown;
+    return (buf.f_flag & ST_LOCAL) ? false : true;
+#elif defined(MNT_LOCAL)
+    struct statfs buf {};
+    if (statfs(narrow.c_str(), &buf) < 0) return dir_remoteness_t::unknown;
+    return (buf.f_flags & MNT_LOCAL) ? dir_remoteness_t::local : dir_remoteness_t::remote;
+#else
+    return dir_remoteness_t::unknown;
+#endif
 }
 
 wcstring_list_t path_get_paths(const wcstring &cmd, const environment_t &vars) {
@@ -126,11 +170,8 @@ wcstring_list_t path_get_paths(const wcstring &cmd, const environment_t &vars) {
     return paths;
 }
 
-maybe_t<wcstring> path_get_cdpath(const wcstring &dir, const wcstring &wd,
+wcstring_list_t path_apply_cdpath(const wcstring &dir, const wcstring &wd,
                                   const environment_t &env_vars) {
-    int err = ENOENT;
-    if (dir.empty()) return none();
-    assert(!wd.empty() && wd.back() == L'/');
     wcstring_list_t paths;
     if (dir.at(0) == L'/') {
         // Absolute path.
@@ -170,11 +211,21 @@ maybe_t<wcstring> path_get_cdpath(const wcstring &dir, const wcstring &wd,
         }
     }
 
-    for (const wcstring &dir : paths) {
+    return paths;
+}
+
+maybe_t<wcstring> path_get_cdpath(const wcstring &dir, const wcstring &wd,
+                                  const environment_t &env_vars) {
+    int err = ENOENT;
+    if (dir.empty()) return none();
+    assert(!wd.empty() && wd.back() == L'/');
+    auto paths = path_apply_cdpath(dir, wd, env_vars);
+
+    for (const wcstring &a_dir : paths) {
         struct stat buf;
-        if (wstat(dir, &buf) == 0) {
+        if (wstat(a_dir, &buf) == 0) {
             if (S_ISDIR(buf.st_mode)) {
-                return dir;
+                return a_dir;
             }
             err = ENOTDIR;
         }
@@ -281,14 +332,17 @@ static int create_directory(const wcstring &d) {
     return ok ? 0 : -1;
 }
 
+namespace {
 /// The following type wraps up a user's "base" directories, corresponding (conceptually if not
 /// actually) to XDG spec.
 struct base_directory_t {
-    wcstring path{};       /// the path where we attempted to create the directory.
-    bool success{false};   /// whether creating the directory succeeded.
-    int err{0};            /// the error code if creating the directory failed.
+    wcstring path{};  /// the path where we attempted to create the directory.
+    dir_remoteness_t remoteness{dir_remoteness_t::unknown};  // whether the dir is remote
+    int err{0};  /// the error code if creating the directory failed, or 0 on success.
+    bool success() const { return err == 0; }
     bool used_xdg{false};  /// whether an XDG variable was used in resolving the directory.
 };
+}  // namespace
 
 /// Attempt to get a base directory, creating it if necessary. If a variable named \p xdg_var is
 /// set, use that directory; otherwise use the path \p non_xdg_homepath rooted in $HOME. \return the
@@ -312,8 +366,15 @@ static base_directory_t make_base_directory(const wcstring &xdg_var,
     }
 
     errno = 0;
-    result.success = !result.path.empty() && create_directory(result.path) != -1;
-    result.err = errno;
+    if (result.path.empty()) {
+        result.err = ENOENT;
+    } else if (create_directory(result.path) < 0) {
+        result.err = errno;
+    } else {
+        result.err = 0;
+        // Need to append a trailing slash to check the contents of the directory, not its parent.
+        result.remoteness = path_remoteness(result.path + L'/');
+    }
     return result;
 }
 
@@ -327,33 +388,43 @@ static const base_directory_t &get_config_directory() {
     return s_dir;
 }
 
-void path_emit_config_directory_errors(env_stack_t &vars) {
+void path_emit_config_directory_messages(env_stack_t &vars) {
     const auto &data = get_data_directory();
-    if (!data.success) {
-        maybe_issue_path_warning(L"data", _(L"Your history will not be saved."), data.used_xdg,
+    if (!data.success()) {
+        maybe_issue_path_warning(L"data", _(L"can not save history"), data.used_xdg,
                                  L"XDG_DATA_HOME", data.path, data.err, vars);
     }
+    if (data.remoteness == dir_remoteness_t::remote) {
+        FLOG(path, "data path appears to be on a network volume");
+    }
 
     const auto &config = get_config_directory();
-    if (!config.success) {
-        maybe_issue_path_warning(L"config", _(L"Your personal settings will not be saved."),
+    if (!config.success()) {
+        maybe_issue_path_warning(L"config", _(L"can not save universal variables or functions"),
                                  config.used_xdg, L"XDG_CONFIG_HOME", config.path, config.err,
                                  vars);
     }
+    if (config.remoteness == dir_remoteness_t::remote) {
+        FLOG(path, "config path appears to be on a network volume");
+    }
 }
 
 bool path_get_config(wcstring &path) {
     const auto &dir = get_config_directory();
-    path = dir.success ? dir.path : L"";
-    return dir.success;
+    path = dir.success() ? dir.path : L"";
+    return dir.success();
 }
 
 bool path_get_data(wcstring &path) {
     const auto &dir = get_data_directory();
-    path = dir.success ? dir.path : L"";
-    return dir.success;
+    path = dir.success() ? dir.path : L"";
+    return dir.success();
 }
 
+dir_remoteness_t path_get_data_remoteness() { return get_data_directory().remoteness; }
+
+dir_remoteness_t path_get_config_remoteness() { return get_config_directory().remoteness; }
+
 void path_make_canonical(wcstring &path) {
     // Ignore trailing slashes, unless it's the first character.
     size_t len = path.size();
