@@ -98,7 +98,7 @@ tok_t::tok_t(token_type_t type) : type(type) {}
 /// Tests if this character can be a part of a string. The redirect ^ is allowed unless it's the
 /// first character. Hash (#) starts a comment if it's the first character in a token; otherwise it
 /// is considered a string character. See issue #953.
-static bool tok_is_string_character(wchar_t c, bool is_first) {
+static bool tok_is_string_character(wchar_t c, bool is_first, maybe_t<wchar_t> next) {
     switch (c) {
         case L'\0':
         case L' ':
@@ -108,11 +108,16 @@ static bool tok_is_string_character(wchar_t c, bool is_first) {
         case L';':
         case L'\r':
         case L'<':
-        case L'>':
-        case L'&': {
+        case L'>': {
             // Unconditional separators.
             return false;
         }
+        case L'&': {
+            if (!feature_test(features_t::ampersand_nobg_in_token)) return false;
+            bool next_is_string = next && tok_is_string_character(*next, false, none());
+            // Unlike in other shells, '&' is not special if followed by a string character.
+            return next_is_string;
+        }
         case L'^': {
             // Conditional separator.
             return !caret_redirs() || !is_first;
@@ -145,9 +150,24 @@ tok_t tokenizer_t::read_string() {
     std::vector<int> paran_offsets;
     std::vector<int> brace_offsets;
     std::vector<char> expecting;
+    std::vector<size_t> quoted_cmdsubs;
     int slice_offset = 0;
     const wchar_t *const buff_start = this->token_cursor;
     bool is_first = true;
+    bool is_token_begin = true;
+
+    auto process_opening_quote = [&](wchar_t quote) -> const wchar_t * {
+        const wchar_t *end = quote_end(this->token_cursor, quote);
+        if (end) {
+            if (*end == L'$') quoted_cmdsubs.push_back(paran_offsets.size());
+            this->token_cursor = end;
+            return nullptr;
+        } else {
+            const wchar_t *error_loc = this->token_cursor;
+            this->token_cursor += std::wcslen(this->token_cursor);
+            return error_loc;
+        }
+    };
 
     while (true) {
         wchar_t c = *this->token_cursor;
@@ -173,6 +193,8 @@ tok_t tokenizer_t::read_string() {
         // has been explicitly ignored (escaped).
         else if (c == L'\\') {
             mode |= tok_modes::char_escape;
+        } else if (c == L'#' && is_token_begin) {
+            this->token_cursor = comment_end(this->token_cursor) - 1;
         } else if (c == L'(') {
             paran_offsets.push_back(this->token_cursor - this->start);
             expecting.push_back(L')');
@@ -195,6 +217,19 @@ tok_t tokenizer_t::read_string() {
                 mode &= ~(tok_modes::subshell);
             }
             expecting.pop_back();
+            // Check if the ) completed a quoted command substitution.
+            if (!quoted_cmdsubs.empty() && quoted_cmdsubs.back() == paran_offsets.size()) {
+                quoted_cmdsubs.pop_back();
+                // The "$(" part of a quoted command substitution closes double quotes. To keep
+                // quotes balanced, act as if there was an invisible double quote after the ")".
+                if (const wchar_t *error_loc = process_opening_quote(L'"')) {
+                    if (!this->accept_unfinished) {
+                        return this->call_error(tokenizer_error_t::unterminated_quote, buff_start,
+                                                error_loc);
+                    }
+                    break;
+                }
+            }
         } else if (c == L'}') {
             if (!expecting.empty() && expecting.back() == L')') {
                 return this->call_error(tokenizer_error_t::expected_pclose_found_bclose,
@@ -225,19 +260,15 @@ tok_t tokenizer_t::read_string() {
         else if (c == L']' && ((mode & tok_modes::array_brackets) == tok_modes::array_brackets)) {
             mode &= ~(tok_modes::array_brackets);
         } else if (c == L'\'' || c == L'"') {
-            const wchar_t *end = quote_end(this->token_cursor);
-            if (end) {
-                this->token_cursor = end;
-            } else {
-                const wchar_t *error_loc = this->token_cursor;
-                this->token_cursor += std::wcslen(this->token_cursor);
-                if ((!this->accept_unfinished)) {
+            if (const wchar_t *error_loc = process_opening_quote(c)) {
+                if (!this->accept_unfinished) {
                     return this->call_error(tokenizer_error_t::unterminated_quote, buff_start,
                                             error_loc);
                 }
                 break;
             }
-        } else if (mode == tok_modes::regular_text && !tok_is_string_character(c, is_first)) {
+        } else if (mode == tok_modes::regular_text &&
+                   !tok_is_string_character(c, is_first, this->token_cursor[1])) {
             break;
         }
 
@@ -250,11 +281,12 @@ tok_t tokenizer_t::read_string() {
         FLOGF(error, msg.c_str(), c, c, int(mode_begin), int(mode));
 #endif
 
-        this->token_cursor++;
+        is_token_begin = is_token_delimiter(this->token_cursor[0], is_first, this->token_cursor[1]);
         is_first = false;
+        this->token_cursor++;
     }
 
-    if ((!this->accept_unfinished) && (mode != tok_modes::regular_text)) {
+    if (!this->accept_unfinished && (mode != tok_modes::regular_text)) {
         if (mode & tok_modes::char_escape) {
             return this->call_error(tokenizer_error_t::unterminated_escape, buff_start,
                                     this->token_cursor - 1, 1);
@@ -512,8 +544,7 @@ maybe_t<tok_t> tokenizer_t::next() {
     while (*this->token_cursor == L'#') {
         // We have a comment, walk over the comment.
         const wchar_t *comment_start = this->token_cursor;
-        while (this->token_cursor[0] != L'\n' && this->token_cursor[0] != L'\0')
-            this->token_cursor++;
+        this->token_cursor = comment_end(this->token_cursor);
         size_t comment_len = this->token_cursor - comment_start;
 
         // If we are going to continue after the comment, skip any trailing newline.
@@ -651,6 +682,10 @@ maybe_t<tok_t> tokenizer_t::next() {
     return result;
 }
 
+bool is_token_delimiter(wchar_t c, bool is_first, maybe_t<wchar_t> next) {
+    return c == L'(' || !tok_is_string_character(c, is_first, next);
+}
+
 wcstring tok_first(const wcstring &str) {
     tokenizer_t t(str.c_str(), 0);
     if (auto token = t.next()) {
@@ -742,7 +777,7 @@ bool move_word_state_machine_t::is_path_component_character(wchar_t c) {
     // Always treat separators as first. All this does is ensure that we treat ^ as a string
     // character instead of as stderr redirection, which I hypothesize is usually what is
     // desired.
-    return tok_is_string_character(c, true) && !std::wcschr(L"/={,}'\":@", c);
+    return tok_is_string_character(c, true, none()) && !std::wcschr(L"/={,}'\":@", c);
 }
 
 bool move_word_state_machine_t::consume_char_path_components(wchar_t c) {
@@ -752,18 +787,22 @@ bool move_word_state_machine_t::consume_char_path_components(wchar_t c) {
         s_separator,
         s_slash,
         s_path_component_characters,
+        s_initial_separator,
         s_end
     };
 
-    // std::fwprintf(stdout, L"state %d, consume '%lc'\n", state, c);
     bool consumed = false;
     while (state != s_end && !consumed) {
         switch (state) {
             case s_initial_punctuation: {
-                if (!is_path_component_character(c)) {
-                    consumed = true;
+                if (!is_path_component_character(c) && !iswspace(c)) {
+                    state = s_initial_separator;
+                } else {
+                    if (!is_path_component_character(c)) {
+                        consumed = true;
+                    }
+                    state = s_whitespace;
                 }
-                state = s_whitespace;
                 break;
             }
             case s_whitespace: {
@@ -800,6 +839,17 @@ bool move_word_state_machine_t::consume_char_path_components(wchar_t c) {
                 }
                 break;
             }
+            case s_initial_separator: {
+                if (is_path_component_character(c)) {
+                    consumed = true;
+                    state = s_path_component_characters;
+                } else if (iswspace(c)) {
+                    state = s_end;
+                } else {
+                    consumed = true;
+                }
+                break;
+            }
             case s_end:
             default: {
                 break;
