@@ -2,6 +2,8 @@
 #define FISH_NO_ISW_WRAPPERS
 #include "config.h"
 
+#include "wutil.h"  // IWYU pragma: keep
+
 #include <dirent.h>
 #include <errno.h>
 #include <fcntl.h>
@@ -9,35 +11,28 @@
 #include <limits.h>
 #include <stdio.h>
 #include <stdlib.h>
-#include <sys/stat.h>
-
-#include <cstring>
 #include <sys/mount.h>
+#include <sys/stat.h>
 #include <sys/statvfs.h>
 #include <sys/types.h>
 #include <unistd.h>
 #include <wctype.h>
 
 #include <atomic>
+#include <cstring>
 #include <cwchar>
 #include <string>
 #include <unordered_map>
 
 #include "common.h"
 #include "fallback.h"  // IWYU pragma: keep
+#include "fds.h"
 #include "flog.h"
 #include "wcstringutil.h"
-#include "wutil.h"  // IWYU pragma: keep
 
 using cstring = std::string;
 
-const file_id_t kInvalidFileID = {static_cast<dev_t>(-1LL),
-                                  static_cast<ino_t>(-1LL),
-                                  static_cast<uint64_t>(-1LL),
-                                  -1,
-                                  -1,
-                                  -1,
-                                  -1};
+const file_id_t kInvalidFileID{};
 
 /// Map used as cache by wgettext.
 static owning_lock<std::unordered_map<wcstring, wcstring>> wgettext_map;
@@ -99,7 +94,7 @@ bool wreaddir(DIR *dir, wcstring &out_name) {
     return true;
 }
 
-bool wreaddir_for_dirs(DIR *dir, wcstring *out_name) {
+bool readdir_for_dirs(DIR *dir, std::string *out_name) {
     struct dirent *result = nullptr;
     while (!result) {
         result = readdir(dir);
@@ -123,7 +118,7 @@ bool wreaddir_for_dirs(DIR *dir, wcstring *out_name) {
     }
 
     if (result && out_name) {
-        *out_name = str2wcstring(result->d_name);
+        *out_name = result->d_name;
     }
     return result != nullptr;
 }
@@ -139,7 +134,6 @@ wcstring wgetcwd() {
     return wcstring();
 }
 
-
 DIR *wopendir(const wcstring &name) {
     const cstring tmp = wcs2string(name);
     return opendir(tmp.c_str());
@@ -209,64 +203,27 @@ int make_fd_blocking(int fd) {
     return err == -1 ? errno : 0;
 }
 
-static inline void safe_append(char *buffer, const char *s, size_t buffsize) {
-    std::strncat(buffer, s, buffsize - std::strlen(buffer) - 1);
-}
-
-// In general, strerror is not async-safe, and therefore we cannot use it directly. So instead we
-// have to grub through sys_nerr and sys_errlist directly On GNU toolchain, this will produce a
-// deprecation warning from the linker (!!), which appears impossible to suppress!
-const char *safe_strerror(int err) {
-#if defined(__UCLIBC__)
-    // uClibc does not have sys_errlist, however, its strerror is believed to be async-safe.
-    // See issue #808.
-    return std::strerror(err);
-#elif defined(HAVE__SYS__ERRS) || defined(HAVE_SYS_ERRLIST)
-#ifdef HAVE_SYS_ERRLIST
-    if (err >= 0 && err < sys_nerr && sys_errlist[err] != nullptr) {
-        return sys_errlist[err];
+maybe_t<wcstring> wreadlink(const wcstring &file_name) {
+    struct stat buf;
+    if (lwstat(file_name, &buf) == -1) {
+        return none();
     }
-#elif defined(HAVE__SYS__ERRS)
-    extern const char _sys_errs[];
-    extern const int _sys_index[];
-    extern int _sys_num_err;
-
-    if (err >= 0 && err < _sys_num_err) {
-        return &_sys_errs[_sys_index[err]];
+    ssize_t bufsize = buf.st_size + 1;
+    char target_buf[bufsize];
+    const std::string tmp = wcs2string(file_name);
+    ssize_t nbytes = readlink(tmp.c_str(), target_buf, bufsize);
+    if (nbytes == -1) {
+        wperror(L"readlink");
+        return none();
     }
-#endif  // either HAVE__SYS__ERRS or HAVE_SYS_ERRLIST
-#endif  // defined(HAVE__SYS__ERRS) || defined(HAVE_SYS_ERRLIST)
-
-    int saved_err = errno;
-    static char buff[384];  // use a shared buffer for this case
-    char errnum_buff[64];
-    format_long_safe(errnum_buff, err);
-
-    buff[0] = '\0';
-    safe_append(buff, "unknown error (errno was ", sizeof buff);
-    safe_append(buff, errnum_buff, sizeof buff);
-    safe_append(buff, ")", sizeof buff);
-
-    errno = saved_err;
-    return buff;
-}
-
-void safe_perror(const char *message) {
-    // Note we cannot use strerror, because on Linux it uses gettext, which is not safe.
-    int err = errno;
-
-    char buff[384];
-    buff[0] = '\0';
-
-    if (message) {
-        safe_append(buff, message, sizeof buff);
-        safe_append(buff, ": ", sizeof buff);
+    // The link might have been modified after our call to lstat.  If the link now points to a path
+    // that's longer than the original one, we can't read everything in our buffer.  Simply give
+    // up. We don't need to report an error since our only caller will already fall back to ENOENT.
+    if (nbytes == bufsize) {
+        return none();
     }
-    safe_append(buff, safe_strerror(err), sizeof buff);
-    safe_append(buff, "\n", sizeof buff);
 
-    ignore_result(write(STDERR_FILENO, buff, std::strlen(buff)));
-    errno = err;
+    return str2wcstring(target_buf, nbytes);
 }
 
 /// Wide character realpath. The last path component does not need to be valid. If an error occurs,
@@ -347,10 +304,12 @@ wcstring normalize_path(const wcstring &path, bool allow_leading_double_slashes)
         }
     }
     wcstring result = join_strings(new_comps, sep);
-    // Prepend one or two leading slashes.
-    // Two slashes are preserved. Three+ slashes are collapsed to one. (!)
-    result.insert(0, allow_leading_double_slashes && leading_slashes > 2 ? 1 : leading_slashes,
-                  sep);
+    // If we don't allow leading double slashes, collapse them to 1 if there are any.
+    int numslashes = leading_slashes > 0 ? 1 : 0;
+    // If we do, prepend one or two leading slashes.
+    // Yes, three+ slashes are collapsed to one. (!)
+    if (allow_leading_double_slashes && leading_slashes == 2) numslashes = 2;
+    result.insert(0, numslashes, sep);
     // Ensure ./ normalizes to . and not empty.
     if (result.empty()) result.push_back(L'.');
     return result;
@@ -505,7 +464,7 @@ ssize_t wwrite_to_fd(const wchar_t *input, size_t input_len, int fd) {
             ssize_t samt = write(fd, cursor, remaining);
             if (samt < 0) return false;
             total_written += samt;
-            size_t amt = static_cast<size_t>(samt);
+            auto amt = static_cast<size_t>(samt);
             assert(amt <= remaining && "Wrote more than requested");
             remaining -= amt;
             cursor += amt;
@@ -580,6 +539,21 @@ locale_t fish_c_locale() {
     return loc;
 }
 
+static bool fish_numeric_locale_is_valid = false;
+
+void fish_invalidate_numeric_locale() { fish_numeric_locale_is_valid = false; }
+
+locale_t fish_numeric_locale() {
+    // The current locale, except LC_NUMERIC isn't forced to C.
+    static locale_t loc = nullptr;
+    if (!fish_numeric_locale_is_valid) {
+        if (loc != nullptr) freelocale(loc);
+        auto cur = duplocale(LC_GLOBAL_LOCALE);
+        loc = newlocale(LC_NUMERIC_MASK, "", cur);
+        fish_numeric_locale_is_valid = true;
+    }
+    return loc;
+}
 /// Like fish_wcstol(), but fails on a value outside the range of an int.
 ///
 /// This is needed because BSD and GNU implementations differ in several ways that make it really
@@ -717,14 +691,14 @@ unsigned long long fish_wcstoull(const wchar_t *str, const wchar_t **endptr, int
 /// Like wcstod(), but wcstod() is enormously expensive on some platforms so this tries to have a
 /// fast path.
 double fish_wcstod(const wchar_t *str, wchar_t **endptr) {
+    // We can ignore the locale because we use LC_NUMERIC=C!
     // The "fast path." If we're all ASCII and we fit inline, use strtod().
     char narrow[128];
     size_t len = std::wcslen(str);
     size_t len_plus_0 = 1 + len;
-    auto is_digit = [](wchar_t c) { return '0' <= c && c <= '9'; };
-    if (len_plus_0 <= sizeof narrow && std::all_of(str, str + len, is_digit)) {
+    auto is_ascii = [](wchar_t c) { return 0 <= c && c <= 127; };
+    if (len_plus_0 <= sizeof narrow && std::all_of(str, str + len, is_ascii)) {
         // Fast path. Copy the string into a local buffer and run strtod() on it.
-        // We can ignore the locale-taking version because we are limited to ASCII digits.
         std::copy(str, str + len_plus_0, narrow);
         char *narrow_endptr = nullptr;
         double ret = strtod(narrow, endptr ? &narrow_endptr : nullptr);
@@ -734,7 +708,7 @@ double fish_wcstod(const wchar_t *str, wchar_t **endptr) {
         }
         return ret;
     }
-    return wcstod_l(str, endptr, fish_c_locale());
+    return std::wcstod(str, endptr);
 }
 
 file_id_t file_id_t::from_stat(const struct stat &buf) {
@@ -771,6 +745,8 @@ file_id_t file_id_for_fd(int fd) {
     return result;
 }
 
+file_id_t file_id_for_fd(const autoclose_fd_t &fd) { return file_id_for_fd(fd.fd()); }
+
 file_id_t file_id_for_path(const wcstring &path) {
     file_id_t result = kInvalidFileID;
     struct stat buf = {};
@@ -793,8 +769,21 @@ bool file_id_t::operator==(const file_id_t &rhs) const { return this->compare_fi
 
 bool file_id_t::operator!=(const file_id_t &rhs) const { return !(*this == rhs); }
 
+wcstring file_id_t::dump() const {
+    using llong = long long;
+    wcstring result;
+    append_format(result, L"     device: %lld\n", llong(device));
+    append_format(result, L"      inode: %lld\n", llong(inode));
+    append_format(result, L"       size: %lld\n", llong(size));
+    append_format(result, L"     change: %lld\n", llong(change_seconds));
+    append_format(result, L"change_nano: %lld\n", llong(change_nanoseconds));
+    append_format(result, L"        mod: %lld\n", llong(mod_seconds));
+    append_format(result, L"   mod_nano: %lld", llong(mod_nanoseconds));
+    return result;
+}
+
 template <typename T>
-int compare(T a, T b) {
+static int compare(T a, T b) {
     if (a < b) {
         return -1;
     } else if (a > b) {
