@@ -19,6 +19,8 @@
 #include "common.h"
 #include "maybe.h"
 
+class autoclose_fd_t;
+
 /// Wide character version of opendir(). Note that opendir() is guaranteed to set close-on-exec by
 /// POSIX (hooray).
 DIR *wopendir(const wcstring &name);
@@ -38,15 +40,12 @@ int wunlink(const wcstring &file_name);
 /// Wide character version of perror().
 void wperror(const wchar_t *s);
 
-/// Async-safe version of perror().
-void safe_perror(const char *message);
-
-/// Async-safe version of std::strerror().
-const char *safe_strerror(int err);
-
 /// Wide character version of getcwd().
 wcstring wgetcwd();
 
+/// Wide character version of readlink().
+maybe_t<wcstring> wreadlink(const wcstring &file_name);
+
 /// Wide character version of realpath function.
 /// \returns the canonicalized path, or none if the path is invalid.
 maybe_t<wcstring> wrealpath(const wcstring &pathname);
@@ -72,7 +71,7 @@ bool wreaddir_resolving(DIR *dir, const std::wstring &dir_path, wcstring &out_na
 /// (i.e. the file is a symlink), then return it. Note that this does not guarantee that everything
 /// returned is a directory, it's just an optimization for cases where we would check for
 /// directories anyways.
-bool wreaddir_for_dirs(DIR *dir, wcstring *out_name);
+bool readdir_for_dirs(DIR *dir, std::string *out_name);
 
 /// Wide character version of dirname().
 std::wstring wdirname(std::wstring path);
@@ -125,6 +124,9 @@ int fish_wcswidth(const wcstring &str);
 // returns an immortal locale_t corresponding to the C locale.
 locale_t fish_c_locale();
 
+void fish_invalidate_numeric_locale();
+locale_t fish_numeric_locale();
+
 int fish_wcstoi(const wchar_t *str, const wchar_t **endptr = nullptr, int base = 10);
 long fish_wcstol(const wchar_t *str, const wchar_t **endptr = nullptr, int base = 10);
 long long fish_wcstoll(const wchar_t *str, const wchar_t **endptr = nullptr, int base = 10);
@@ -137,13 +139,15 @@ double fish_wcstod(const wchar_t *str, wchar_t **endptr);
 /// seems to aggressively re-use inodes, so it cannot determine if a file has been deleted (ABA
 /// problem). Therefore we include richer information.
 struct file_id_t {
-    dev_t device;
-    ino_t inode;
-    uint64_t size;
-    time_t change_seconds;
-    long change_nanoseconds;
-    time_t mod_seconds;
-    long mod_nanoseconds;
+    dev_t device{static_cast<dev_t>(-1LL)};
+    ino_t inode{static_cast<ino_t>(-1LL)};
+    uint64_t size{static_cast<uint64_t>(-1LL)};
+    time_t change_seconds{-1};
+    long change_nanoseconds{-1};
+    time_t mod_seconds{-1};
+    long mod_nanoseconds{-1};
+
+    constexpr file_id_t() = default;
 
     bool operator==(const file_id_t &rhs) const;
     bool operator!=(const file_id_t &rhs) const;
@@ -153,6 +157,8 @@ struct file_id_t {
 
     static file_id_t from_stat(const struct stat &buf);
 
+    wcstring dump() const;
+
    private:
     int compare_file_id(const file_id_t &rhs) const;
 };
@@ -182,6 +188,7 @@ struct hash<file_id_t> {
 #endif
 
 file_id_t file_id_for_fd(int fd);
+file_id_t file_id_for_fd(const autoclose_fd_t &fd);
 file_id_t file_id_for_path(const wcstring &path);
 file_id_t file_id_for_path(const std::string &path);
 
